extends Node
class_name CubeSpawner

@export var left_hand_pinch_detector : PinchDetector
@export var right_hand_pinch_detector : PinchDetector

@export var min_cube_size : float = 0.01
@export var max_cube_size : float = 0.2

@export var max_cubes : int = 10

@export var pinch_distance : float = 0.1

@export var set_force : bool = true
@export var set_torque : bool = true

@export var cube_spawn_material : StandardMaterial3D

var size_offset : float = 0.04
var min_distance : float = min_cube_size + size_offset

var both_pinched : bool = false
var spawning_cube : Node3D = null

var cubes : Array = []
var cube_resource : Resource = preload("res://scenes/Cube.tscn")

var spawn_position : Vector3 = Vector3.ZERO
var spawn_size : Vector3 = Vector3.ZERO
var spawn_basis : Basis = Basis.IDENTITY

var prev_cube_transform : Transform3D 
var latest_colour : String = "E40303"

var colours : Dictionary = {
	"cyan": "#00F1FF",
	"blue": "#0161E8",
	"bluer": "#290CFF",
	"purple": "#9B00E8",
	"pink": "#FF019A"
}

var rng : RandomNumberGenerator = RandomNumberGenerator.new()

func _physics_process(delta):
	if cubes.size() > max_cubes:
		var to_remove : Array = cubes.slice(0, -max_cubes)
		for cube in to_remove:
			cube.queue_free()
		cubes = cubes.slice(-max_cubes)
		
	if left_hand_pinch_detector.is_pinched && right_hand_pinch_detector.is_pinched:
		var left_hand : UltraleapHand = left_hand_pinch_detector.hand
		var right_hand : UltraleapHand = right_hand_pinch_detector.hand
		
		#var dist : float = left_hand.thumb.distal.next_joint.distance_to(right_hand.thumb.distal.next_joint)
		var dist : float = left_hand_pinch_detector.pinch_transform.origin.distance_to(
			right_hand_pinch_detector.pinch_transform.origin
		)
		
		# Detect when they just got pinched
		if !both_pinched && dist <= pinch_distance:
			spawning_cube = CSGBox3D.new()
			get_tree().get_root().add_child(spawning_cube)
			spawning_cube.owner = get_tree().get_root()
			spawning_cube.material = cube_spawn_material
			spawning_cube.material.setup_local_to_scene()
			
			var i : int = rng.randi_range(0, colours.keys().size() - 1)
			latest_colour = colours[colours.keys()[i]]
			spawning_cube.material.albedo_color = latest_colour
			spawning_cube.material.emission = latest_colour

		both_pinched = true

		if spawning_cube == null:
			return
			
		# Might move this line around
		prev_cube_transform = Transform3D(spawn_basis, spawn_position)
		
		# Get spawn size and position to reuse later and to display now on the model
		
		var size : float = clamp(dist - size_offset, min_cube_size, max_cube_size)
		
		spawn_size = Vector3(size, size, size)
		#spawn_position = (right_hand.thumb.distal.next_joint + left_hand.thumb.distal.next_joint) / 2
		spawn_position = left_hand_pinch_detector.pinch_transform.interpolate_with(
			right_hand_pinch_detector.pinch_transform, 0.5
		).origin

		#var line : Vector3 = right_hand.thumb.distal.next_joint - left_hand.thumb.distal.next_joint
		var line : Vector3 = right_hand_pinch_detector.pinch_transform.origin - left_hand_pinch_detector.pinch_transform.origin

		# Meh... Calculating the euler angles. Feels like it could be better...
		var z_angle : float = -atan2(line.y, line.x)
		var y_angle : float = -atan2(line.z, line.x)
		
		spawn_basis = Basis.IDENTITY
		spawn_basis = spawn_basis.rotated(Vector3.FORWARD, z_angle).rotated(Vector3.UP, y_angle)

		spawning_cube.transform = Transform3D(spawn_basis.scaled(spawn_size), spawn_position)
		
	else:
		
		if both_pinched && spawning_cube != null:
			# Drop the cube, so we instantiate a new one with physics and remove the model
			var new_cube = cube_resource.instantiate()

			var cube_transform : Transform3D = Transform3D(spawn_basis, spawn_position)
			new_cube.set_global_transform(cube_transform)
			cube_transform.basis = cube_transform.basis.scaled(spawn_size)
			new_cube.get_node("CSGBox3D").size = cube_transform.basis.get_scale()
			new_cube.get_node("CollisionShape3D").shape.size = cube_transform.basis.get_scale()
			new_cube.get_node("CSGBox3D").material.albedo_color = latest_colour
			new_cube.get_node("CSGBox3D").material.emission = latest_colour
			
			# Seems better if I set inertia to zero (recomputes propa' inertia ?)
			new_cube.inertia = Vector3.ZERO
			
			# Eventually we'll want to find the pivot point on the line between the pinches so we can calculate a better torque ?
			
			var torque : Vector3 = ((new_cube.get_inverse_inertia_tensor() * prev_cube_transform.basis.transposed() * cube_transform.basis).get_euler() / delta) * (PI / 180)
			var force : Vector3 = (spawn_position - prev_cube_transform.origin) / delta
			
			if set_torque:
				new_cube.apply_torque(torque)
			if set_force:
				new_cube.apply_impulse(force)
			
			get_tree().get_root().add_child(new_cube)
			new_cube.owner = get_tree().get_root()
			cubes.append(new_cube)
			spawning_cube.queue_free()

		both_pinched = false
